"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const config_1 = require("./config");
const configFileWatcher_1 = require("./configFileWatcher");
const runner_1 = require("./runner");
const settings_1 = require("./settings");
const failures_1 = require("./runner/failures");
class FailureMap {
    constructor() {
        this._map = new Map();
    }
    get(start, end) {
        return this._map.get(this.key(start, end));
    }
    set(start, end, failure) {
        this._map.set(this.key(start, end), failure);
    }
    values() {
        return this._map.values();
    }
    // key to identify a rule failure
    key(start, end) {
        return `[${start},${end}]`;
    }
}
class TSLintPlugin {
    constructor(ts, logger, project, config) {
        this.ts = ts;
        this.logger = logger;
        this.project = project;
        this.codeFixActions = new Map();
        this.logger.info('loaded');
        this.config = settings_1.loadSettingsFromPluginConfig(config, project.getCurrentDirectory());
        this.runner = new runner_1.TsLintRunner(message => { this.logger.info(message); });
        // Watch config file for changes
        if (project instanceof ts.server.ConfiguredProject && ts.sys.watchFile) {
            const configFile = project.getConfigFilePath();
            this.logger.info(`Found configured project: ${configFile}`);
            ts.sys.watchFile(configFile, (_fileName, eventKind) => {
                if (eventKind !== ts.FileWatcherEventKind.Changed) {
                    return;
                }
                this.logger.info('TSConfig file changed');
                const newConfig = settings_1.loadSettingsFromTsConfig(ts, configFile, this.project.getCurrentDirectory());
                if (!newConfig) {
                    this.logger.info(`Could not read new config`);
                    return;
                }
                this.logger.info(`Updating config settings: ${JSON.stringify(newConfig)}`);
                this.config = newConfig;
                this.project.refreshDiagnostics();
            });
        }
        this.configFileWatcher = new configFileWatcher_1.ConfigFileWatcher(ts, filePath => {
            this.logger.info('TSlint file changed');
            this.runner.onConfigFileChange(filePath);
            this.project.refreshDiagnostics();
        });
    }
    decorate(languageService) {
        const oldGetSupportedCodeFixes = this.ts.getSupportedCodeFixes.bind(this.ts);
        this.ts.getSupportedCodeFixes = () => {
            return [
                ...oldGetSupportedCodeFixes(),
                config_1.TSLINT_ERROR_CODE,
            ];
        };
        const oldGetSemanticDiagnostics = languageService.getSemanticDiagnostics.bind(languageService);
        languageService.getSemanticDiagnostics = (fileName) => {
            return this.getSemanticDiagnostics(oldGetSemanticDiagnostics, fileName);
        };
        const getCodeFixesAtPosition = languageService.getCodeFixesAtPosition.bind(languageService);
        languageService.getCodeFixesAtPosition = (fileName, start, end, errorCodes, formatOptions, userPreferences) => {
            return this.getCodeFixesAtPosition(getCodeFixesAtPosition, fileName, start, end, errorCodes, formatOptions, userPreferences);
        };
        return languageService;
    }
    getSemanticDiagnostics(delegate, fileName) {
        const diagnostics = delegate(fileName);
        if (this.config.suppressWhileTypeErrorsPresent && diagnostics.length > 0) {
            return diagnostics;
        }
        try {
            this.logger.info(`Computing tslint semantic diagnostics for '${fileName}'`);
            if (this.codeFixActions.has(fileName)) {
                this.codeFixActions.delete(fileName);
            }
            if (this.config.ignoreDefinitionFiles === true && fileName.endsWith('.d.ts')) {
                return diagnostics;
            }
            let result;
            try { // protect against tslint crashes
                result = this.runner.runTsLint(fileName, this.getProgram(), {
                    configFile: this.config.configFile,
                    ignoreDefinitionFiles: this.config.ignoreDefinitionFiles,
                    jsEnable: this.config.jsEnable,
                });
                if (result.configFilePath) {
                    this.configFileWatcher.ensureWatching(result.configFilePath);
                }
            }
            catch (err) {
                let errorMessage = `unknown error`;
                if (typeof err.message === 'string' || err.message instanceof String) {
                    errorMessage = err.message;
                }
                this.logger.info('tslint error ' + errorMessage);
                return diagnostics;
            }
            const file = this.getProgram().getSourceFile(fileName);
            for (const warning of result.warnings) {
                this.logger.info(`[tslint] ${warning}`);
                diagnostics.push({
                    code: config_1.TSLINT_ERROR_CODE,
                    source: config_1.TSLINT_ERROR_SOURCE,
                    category: this.ts.DiagnosticCategory.Error,
                    file,
                    start: 0,
                    length: 1,
                    messageText: warning,
                });
            }
            const tslintProblems = failures_1.filterProblemsForFile(fileName, result.lintResult.failures);
            for (const problem of tslintProblems) {
                diagnostics.push(this.makeDiagnostic(problem, file));
                this.recordCodeAction(problem, file);
            }
        }
        catch (e) {
            this.logger.info(`tslint-language service error: ${e.toString()}`);
            this.logger.info(`Stack trace: ${e.stack}`);
        }
        return diagnostics;
    }
    getCodeFixesAtPosition(delegate, fileName, start, end, errorCodes, formatOptions, userPreferences) {
        const fixes = delegate(fileName, start, end, errorCodes, formatOptions, userPreferences);
        if (this.config.suppressWhileTypeErrorsPresent && fixes.length > 0) {
            return fixes;
        }
        this.logger.info(`getCodeFixes ${errorCodes[0]}`);
        const documentFixes = this.codeFixActions.get(fileName);
        if (documentFixes) {
            const problem = documentFixes.get(start, end);
            if (problem) {
                const fix = problem.getFix();
                if (fix) {
                    fixes.push(this.getRuleFailureQuickFix(problem, fileName));
                    const fixAll = this.getRuleFailureFixAllQuickFix(problem.getRuleName(), documentFixes, fileName);
                    if (fixAll) {
                        fixes.push(fixAll);
                    }
                }
            }
            fixes.push(this.getFixAllAutoFixableQuickFix(documentFixes, fileName));
            if (problem) {
                fixes.push(this.getDisableRuleQuickFix(problem, fileName, this.getProgram().getSourceFile(fileName)));
            }
        }
        return fixes;
    }
    recordCodeAction(problem, file) {
        let fix;
        // tslint can return a fix with an empty replacements array, these fixes are ignored
        if (problem.getFix && problem.getFix() && !replacementsAreEmpty(problem.getFix())) { // tslint fixes are not available in tslint < 3.17
            fix = problem.getFix(); // createAutoFix(problem, document, problem.getFix());
        }
        if (!fix) {
            return;
        }
        let documentAutoFixes = this.codeFixActions.get(file.fileName);
        if (!documentAutoFixes) {
            documentAutoFixes = new FailureMap();
            this.codeFixActions.set(file.fileName, documentAutoFixes);
        }
        documentAutoFixes.set(problem.getStartPosition().getPosition(), problem.getEndPosition().getPosition(), problem);
    }
    getRuleFailureQuickFix(problem, fileName) {
        return {
            description: `Fix: ${problem.getFailure()}`,
            fixName: '',
            changes: [problemToFileTextChange(problem, fileName)],
        };
    }
    /**
     * Generate a code action that fixes all instances of ruleName.
     */
    getRuleFailureFixAllQuickFix(ruleName, problems, fileName) {
        const changes = [];
        for (const problem of problems.values()) {
            if (problem.getRuleName() === ruleName) {
                changes.push(problemToFileTextChange(problem, fileName));
            }
        }
        // No need for this action if there's only one instance.
        if (changes.length < 2) {
            return undefined;
        }
        return {
            description: `Fix all '${ruleName}'`,
            fixName: '',
            changes,
        };
    }
    getDisableRuleQuickFix(problem, fileName, file) {
        return {
            description: `Disable rule '${problem.getRuleName()}'`,
            fixName: '',
            changes: [{
                    fileName,
                    textChanges: [{
                            newText: `// tslint:disable-next-line: ${problem.getRuleName()}\n`,
                            span: { start: file.getLineStarts()[problem.getStartPosition().getLineAndCharacter().line], length: 0 },
                        }],
                }],
        };
    }
    getFixAllAutoFixableQuickFix(documentFixes, fileName) {
        const allReplacements = failures_1.getNonOverlappingReplacements(Array.from(documentFixes.values()));
        return {
            description: `Fix all auto-fixable tslint failures`,
            fixName: '',
            changes: [{
                    fileName,
                    textChanges: allReplacements.map(convertReplacementToTextChange),
                }],
        };
    }
    getProgram() {
        return this.project.getLanguageService().getProgram();
    }
    makeDiagnostic(problem, file) {
        const message = (problem.getRuleName() !== null)
            ? `${problem.getFailure()} (${problem.getRuleName()})`
            : `${problem.getFailure()}`;
        const category = this.getDiagnosticCategory(problem);
        return {
            file,
            start: problem.getStartPosition().getPosition(),
            length: problem.getEndPosition().getPosition() - problem.getStartPosition().getPosition(),
            messageText: message,
            category,
            source: config_1.TSLINT_ERROR_SOURCE,
            code: config_1.TSLINT_ERROR_CODE,
        };
    }
    getDiagnosticCategory(problem) {
        if (this.config.alwaysShowRuleFailuresAsWarnings === true) {
            return this.ts.DiagnosticCategory.Warning;
        }
        else if (problem.getRuleSeverity && problem.getRuleSeverity() === 'error') {
            // tslint5 supports to assign severities to rules
            return this.ts.DiagnosticCategory.Error;
        }
        return this.ts.DiagnosticCategory.Warning;
    }
}
exports.TSLintPlugin = TSLintPlugin;
function getReplacements(fix) {
    let replacements = null;
    // in tslint4 a Fix has a replacement property with the Replacements
    if (fix.replacements) {
        // tslint4
        replacements = fix.replacements;
    }
    else {
        // in tslint 5 a Fix is a Replacement | Replacement[]
        if (!Array.isArray(fix)) {
            replacements = [fix];
        }
        else {
            replacements = fix;
        }
    }
    return replacements || [];
}
function convertReplacementToTextChange(repl) {
    return {
        newText: repl.text,
        span: { start: repl.start, length: repl.length },
    };
}
function problemToFileTextChange(problem, fileName) {
    const fix = problem.getFix();
    const replacements = getReplacements(fix);
    return {
        fileName,
        textChanges: replacements.map(convertReplacementToTextChange),
    };
}
function replacementsAreEmpty(fix) {
    // in tslint 4 a Fix has a replacement property witht the Replacements
    if (fix.replacements) {
        return fix.replacements.length === 0;
    }
    // tslint 5
    if (Array.isArray(fix)) {
        return fix.length === 0;
    }
    return false;
}
//# sourceMappingURL=plugin.js.map